
import vytools.utils
import vytools.compose
import vytools.definition
import vytools.object
import vytools.composerun
from termcolor import cprint
from pathlib import Path
from vytools.config import CONFIG, ITEMS
import re, json, shutil, os, yaml, subprocess, sys, io
import signal
import cerberus
import logging

SCHEMA = vytools.utils.BASE_SCHEMA.copy()

INHERIT_ITEMS = {
  'tags':{'type':'list', 'required': False, 'schema': {'type': 'string', 'maxlength':64}},
  'expectation':{'type':'boolean', 'required': False},
}

SCHEMA.update(INHERIT_ITEMS)
SCHEMA.update({
  'thingtype':{'type':'string', 'allowed':['episode']},
  'compose':{'type':'string', 'maxlength': 64},
  'object_mods':{'type': 'dict', 'required': False},
  'passed':{'type':'boolean', 'required': False},
  'anchors':{'type': 'dict', 'required': False, 'keysrules': {'type': 'string', 'regex': vytools.compose.KEYSRULES}},
  'repos':{'type':'list', 'required': False, 'schema':{'type':'string','maxlength':1024}},
  'stage_repos':{'type':'list', 'required': False, 'schema':{'type':'string','maxlength':1024}}
})
VALIDATE = cerberus.Validator(SCHEMA)

def parse(name, pth, reponame, items):
  item = {
    'name':name,
    'repo':reponame,
    'thingtype':'episode',
    'path':pth
  }
  type_name = 'episode:' + name
  item['depends_on'] = []
  try:
    content = json.load(io.open(pth, 'r', encoding='utf-8-sig'))
    if 'anchors' not in content: content['anchors'] = {}
      
    for sc in SCHEMA:
      if sc in content and sc not in ['repos']:
        item[sc] = content[sc]

    if vytools.utils._check_add(item['compose'], 'compose', item, items, type_name):
      for anchor_key, anchor_val in content['anchors'].items():
        if anchor_val.startswith('object:'):
          vytools.utils._check_add(anchor_val, 'object', item, items, type_name)
      return vytools.utils._add_item(item, items, VALIDATE)
    else:
      return False
  except Exception as exc:
    logging.error('Failed to parse episode "{n}" at "{p}": {e}'.format(n=name, p=pth, e=exc))
    return False

def find_all(items, contextpaths=None):
  return vytools.utils.search_all(r'(.+)\.episode\.json', parse, items, contextpaths=contextpaths)

def get_anchors(episode, anchor_dict=None):
  if anchor_dict is None: anchor_dict = {}
  anchors = episode.get('anchors',{})
  anchors.update(anchor_dict)
  return anchors

def build(type_name, built, items=None, anchors=None, build_level=0, object_mods=None, eppath=None):
  if items is None: items = ITEMS
  item = items[type_name]
  extracted_anchors = get_anchors(item, anchors)
  rootcompose = item['compose']
  return vytools.compose.build(rootcompose, items=items, anchors=extracted_anchors, built=built, 
                        build_level=build_level, object_mods=object_mods, eppath=eppath)

def get_episode_id(episode):
  return None if '..' in episode['name'] else episode['name'].lower()

def get_episode(episode_name, items=None):
  if items is None: items = ITEMS
  if episode_name not in items:
    vytools.utils.missing_item(episode_name, items)
    return None
  return items[episode_name]

def get_episode_path(episode_name, items=None, jobpath=None):
  ep = get_episode(episode_name, items=items)
  if ep:
    epid = get_episode_id(ep)
    return vytools.composerun.runpath(epid,jobpath=jobpath)
  return None

def artifact_paths(episode_name, items=None, jobpath=None):
  ep = get_episode(episode_name, items=items)
  if ep:
    eppath = get_episode_path(episode_name, items=items, jobpath=jobpath)
    if eppath:
      return vytools.compose.artifact_paths(ep.get('compose',None), items, eppath=eppath)
  return {}

def run(type_name, items=None, anchors=None, clean=False, save=False, object_mods=None, jobpath=None, persist=False):
  if items is None: items = ITEMS
  if type_name not in items:
    vytools.utils.missing_item(type_name, items)
    return False
  episode = items[type_name]
  epid = get_episode_id(episode)
  eppath = vytools.composerun.runpath(epid,jobpath=jobpath)
  if epid is None or eppath is None: return False
  extracted_anchors = get_anchors(episode, anchors)
  results = vytools.composerun.run(epid, episode['compose'], 
    items=items, anchors=extracted_anchors, clean=clean, 
    object_mods=object_mods, jobpath=jobpath,
    dont_track=episode['repo'], persist=persist)

  if results and eppath and os.path.exists(eppath):
    for x in INHERIT_ITEMS:
      if x in episode: results[x] = episode[x]
    with open(os.path.join(eppath, episode['name']+'.episode_.json'),'w') as w2:
      w2.write(json.dumps(results, sort_keys=True, indent=2))
    if save and 'path' in episode:
      # TODO: what should I do to update the episode object loaded in vytools
      with open(episode['path'],'w') as w2:
        w2.write(json.dumps(results, sort_keys=True, indent=2))
  return results




from termcolor import cprint
import os
INFOCOLOR = 'cyan'
SUCCESSCOLOR = 'green'
FAILCOLOR = 'red'
WARNCOLOR = 'yellow'

BUFFER = None
def set_buffer(buffer):
  global BUFFER
  BUFFER = buffer

def printer__(strng, color=None, attrs=None, fp=None):
  if attrs is None: attrs = []
  if fp:
    fp.write(strng+os.linesep)

  if BUFFER is not None:
    BUFFER.append(strng)
  if color or attrs:
    cprint(strng, color=color, attrs=attrs)
  else:
    print(strng)

def print_def(strng, attrs=None, fp=None):
  printer__(strng, attrs=attrs, fp=fp)

def print_plus(strng, fp=None):
  color = WARNCOLOR if strng.endswith('+') else SUCCESSCOLOR
  printer__(strng, color=color, fp=fp)

def print_warn(*argv):
  printer__(' '.join(argv),color=WARNCOLOR)

def print_info(*argv):
  printer__(' '.join(argv),color=INFOCOLOR)

def print_success(*argv):
  printer__(' '.join(argv),color=SUCCESSCOLOR)

def print_fail(*argv):
  printer__(' '.join(argv),color=FAILCOLOR)




import vytools.utils as utils
import vytools.definition
import re, json, io
import cerberus
import logging

SCHEMA = utils.BASE_SCHEMA.copy()
SCHEMA.update({
  'thingtype':{'type':'string', 'allowed':['object']},
  'definition':{'type':'string', 'maxlength': 64},
  'data':{'type': 'dict'}
})
VALIDATE = cerberus.Validator(SCHEMA)

def parse(name, pth, reponame, items):
  item = {
    'name':name,
    'repo':reponame,
    'thingtype':'object',
    'depends_on':[],
    'path':pth
  }
  try:
    content = json.load(io.open(pth, 'r', encoding='utf-8-sig'))
    item['definition'] = content['definition']
    item['data'] = content['data']
  except Exception as exc:
    logging.error('Failed to parse object "{n}" at "{p}": {e}'.format(n=name, p=pth, e=exc))
    return False

  return utils._add_item(item, items, VALIDATE)

def element_length(element):
  return -1 if 'length' not in element else (0 if element['length'] == '?' else int(element['length']))

def check_length(element, obj):
  length = element_length(element)
  if type(obj) == list and length == -1:
    logging.error('Element "{e}" should not be a list'.format(e=element['name']))
    return False
  elif type(obj) != list and length > -1:
    logging.error('Element "{e}" should be a list'.format(e=element['name']))
    return False
  elif type(obj) == list and length > 0 and len(obj) != length:
    logging.error('Element "{e}" should be a list of length {n}'.format(e=element['name'], n=length))
    return False
  return True

def _make_object_mod_sub(object_mods, key):
  keyi = key + '.'
  return {k.replace(key+'.','',1):v for k,v in object_mods.items() if k.startswith(key+'.')}

def _make_object_mod_sub_i(object_mods, key, i):
  object_mods_sub_r = _make_object_mod_sub(object_mods, key + '.$')
  object_mods_sub_i = _make_object_mod_sub(object_mods, key + '.'+str(i))
  object_mods_sub_r.update(object_mods_sub_i)
  return object_mods_sub_r

def expand(data_, definition_=None, items=None, object_mods=None):
  if items is None: items = vytools.ITEMS
  dependencies = []
  if type(data_) == str:
    dependencies += [data_]
    if not data_.startswith('object:'):
      logging.error('Cannot expand the item "{}" since it is not an "object:"'.format(data_))
    elif data_ not in items:
      logging.error('Cannot expand the item "{}" since it is not in the list of known objects'.format(data_))
    else:
      (data__, deps_) = expand(items[data_]['data'], items[data_]['definition'], items, object_mods=object_mods)
      if (data__ is None): return (data__, deps_)
      dependencies += [d for d in deps_ if d not in dependencies]
      return (data__, dependencies)
    return (None, [])

  data = {}
  if object_mods is None: object_mods = {}
  if definition_ is None or definition_ not in items:
    logging.error('The definition {d} does not exist'.format(d=definition_))
    return (None, [])

  definition = items[definition_]
  for el in definition['element']:
    if el['type'] in utils.BASE_DATA_TYPES:
      # TODO make sure these are correct, also, could be list
      rhs = object_mods.get(el['name'], data_.get(el['name'], utils.BASE_DATA_TYPES[el['type']]))
    elif el['type'] in items:
      obj = object_mods.get(el['name'], data_.get(el['name'],{}))
      if type(obj) == str:
        if obj in items:
          defin = items[obj]['definition']
          if defin == el['type']:
            dependencies += [obj]
            obj = items[obj]['data']
          else:
            logging.error('The internal definition of object:\n "{o}" ({d1}) does not match the desired {d2}'.format(o=obj,d1=defin,d2=el['type']))
            return (None, [])
        else:
          logging.error('The object "{o}" does not exist'.format(o=obj))
          return (None, [])
      if type(obj) == list:
        rhs = []
        ocount = 0
        for o in obj:
          dms = _make_object_mod_sub_i(object_mods,el['name'],ocount)
          rhsi,deps_ = expand(o, el['type'], items, object_mods=dms)
          if rhsi is None: return (rhs, deps_)
          dependencies += [d for d in deps_ if d not in dependencies]
          rhs.append(rhsi)
          ocount += 1
      else:
        dms = _make_object_mod_sub(object_mods,el['name'])
        rhs,deps_ = expand(obj, el['type'], items, object_mods=dms)
        if rhs is None: return (rhs, deps_)
        dependencies += [d for d in deps_ if d not in dependencies]
    else:
      return (None, [])
    data[el['name']] = rhs
  return (data, dependencies)

def object_validate(obj, sigtypename, top, items, trace, add_missing_fields):
  success = True
  if sigtypename in utils.BASE_DATA_TYPES:
    pass #TODO validate types
  elif sigtypename in items:
    if type(obj) == str: # This must be a reference to another object
      if obj not in items:
        success = False
        logging.error('Object "{top}" references object "{o}" at {trace} which does not seem to exist'.format(o=obj, top=top, trace=trace))
      elif not obj.startswith('object:'):
        success = False
        logging.error('Object "{top}" references object "{o}" at {trace} which is not of type "object:"'.format(o=obj, top=top, trace=trace))
      else:
        item = items[obj]
        if item['definition'] != sigtypename:
          success = False
          logging.error(('Object definitions dont match in "{top}":\n'+
            '  {top} expects an object of definition type "{d1}" in position {trace}\n'+
            '  but references object "{o}" which is of definition type "{d2}"').format(o=obj, 
                      top=top, d2=item['definition'], d1=sigtypename, trace=trace))
        else:
          items[top]['depends_on'].append(obj)
    elif type(obj) == dict:
      definition = items[sigtypename]
      foundelements = []
      for e in definition['element']:
        key = e['name']
        if key not in obj:
          if key in add_missing_fields['fields']:
            add_missing_fields['added'] = True
            obj[key] = add_missing_fields['fields'][key]
          elif e.get('optional',False):
            pass
          else:
            success = False
            logging.error('Expected element "{e}" in data of {top} at {trace}'.format(e=key, top=top, trace=trace+'.'+key))
        else:
          successi = check_length(e, obj[key])
          if successi:
            if type(obj[key]) == list:
              count=0
              for v in obj[key]:
                success &= object_validate(v, e['type'], top, items, trace+'.'+str(count), add_missing_fields)
                count += 1
            else:
              success &= object_validate(obj[key], e['type'], top, items, trace+'.'+key, add_missing_fields)
          success &= successi
    else:
      success = False
      logging.error('Bad type in object {top}'.format(top=top))
  else:
    success = False
    logging.error('Unknown calibration definition "{s}"'.format(s=sigtypename))
  return success

def find_all(items, add_missing_fields=None, contextpaths=None):
  if add_missing_fields is None: add_missing_fields = {'fields':{},'added':False}
  success = utils.search_all(r'(.+)\.object\.json', parse, items, contextpaths=contextpaths)
  if success:
    for (type_name, item) in items.items():
      if type_name.startswith('object:'):
        (typ, name) = type_name.split(':',1)
        item['depends_on'] = []
        # Add definition to dependency
        if item['definition'] in items:
          item['depends_on'].append(item['definition'])
        else:
          success = False
          logging.error('object "{n}" has an invalid definition {t}'.format(n=name, t=item['definition']))
        success &= object_validate(item['data'], item['definition'], type_name, items, '',add_missing_fields)
        if add_missing_fields['added']:
          add_missing_fields['added'] = False
          with open(item['path'],'r+') as rw:
              d = json.loads(rw.read())
              d['data'] = item['data']
              rw.seek(0)
              rw.write(json.dumps(d,indent=2,sort_keys=True))
              rw.truncate()
  return success



import os, re, logging, subprocess, time, shlex
import vytools.utils as utils
import vytools.printer
from vytools.config import CONFIG, ITEMS
from pathlib import Path
import cerberus
import docker

DOCKER_CLIENT = docker.from_env()

#todo NOT SURE THIS IS THE RIGHT REGEX
COPY_STAGE_A = re.compile(r'^copy[\s]+--from=[\'"]?([\w-]+)',re.I | re.M)
COPY_STAGE_B = re.compile(r'^from[\s]+([\w\-:\.\/]+)',re.I | re.M)
COPY_REPO = re.compile(r'^copy[\s]+[\'"]?([\w-]+)',re.I | re.M)

SCHEMA = utils.BASE_SCHEMA.copy()
SCHEMA.update({
  'thingtype':{'type':'string', 'allowed':['stage']},
  'build_context':{'type':'string', 'maxlength': 1024},
  'secrets':{'type':'list','schema': {'type': 'string', 'maxlength':64}},
  'ssh':{'type':'list','schema': {'type': 'string', 'maxlength':64}},
  'args':{
    'type':'list',
    'schema': {
      'type': 'dict',
      'schema': {
        'key': {'type': 'string', 'maxlength': 64},
        'value': {'type': 'string', 'maxlength': 64}
      }
    }
  }
})

VALIDATE = cerberus.Validator(SCHEMA)

def get_repo_info(name):
  return {'type':'git','source':'bitbucket','account':'autonomoussolutions','name':name}

def line_before_escape(line):
  return re.split(r'([^\\]#)',line)[0]      # All characters before unescaped # (comments)

def remove_quotes(line):
  if line.startswith('"') and line.endswith('"'):
    return line.strip('"')
  elif line.startswith("'") and line.endswith("'"):
    return line.strip("'")
  else:
    return line

def dependencies(stage_text, external_images):
  dependency = []
  args = []
  started = False
  for linex in stage_text.splitlines(): # TODO would be nice to preserve backslash continuation lines as one line
    line = linex.strip()
    if line.startswith('#'): continue         # need this
    line = line_before_escape(line)           # All characters before unescaped # (comments)
    m = re.match(COPY_STAGE_A, line)
    if not m:
      m = re.match(COPY_STAGE_B, line)
      if m: started = True
    if m and 'stage:'+m.group(1) not in dependency:  # COPY/FROM STAGE
      if m.group(1) not in external_images:
        dependency.append('stage:'+m.group(1))  # add dependency
    if started: # For multistage builds only get args AFTER FROM (only these will work)
      m = re.match(r'^[\s]*arg[\s]+(.+)',line,re.I)
      if m:
        argline = line_before_escape(m.group(1)).split('=',1)
        argkey = argline[0].strip()
        argval = argline[1].strip() if len(argline) > 1 else ''
        if argkey not in [a['key'] for a in args]:
          args.append({'key':argkey, 'value':remove_quotes(argval)}) 
  return (dependency, args)

def parse(namein, pth, reponame, items):
  name = namein.lower() # Stages must be unique case insensitively
  stage_text = Path(pth).read_text()
  secrets = []
  ssh = []
  build_context = '.'
  external_images = []
  for line in stage_text.splitlines():
    m = re.search(r'^#vy(.+)', line, re.I | re.M)
    if m:
      lsplit = shlex.split(line)
      if len(lsplit) > 2:
        if lsplit[1].lower() == 'context':
          build_context = lsplit[2]
        elif lsplit[1].lower() == 'source' and lsplit[2] not in external_images:
          external_images.append(lsplit[2])
    m = re.search(r'--mount=type=secret(.+)',line,re.I | re.M)
    if m:
      m = re.search(r'id=([^\s]+)',m.group(1),re.I | re.M)
      if m and m.group(1) not in secrets:
        secrets.append(m.group(1))
    m = re.search(r'--mount=type=ssh(.+)',line,re.I | re.M)
    if m:
      m = re.search(r'id=([^\s]+)',m.group(1),re.I | re.M)
      if m and m.group(1) not in ssh:
        ssh.append(m.group(1))
  type_name = 'stage:'+name
  if os.path.isabs(build_context):
    logging.error('Stage build contexts should be relative paths "{p}"'.format(p=build_context))
    return False
  else:
    (dependency,args) = dependencies(stage_text, external_images)
    try:
      build_context = Path(os.path.join(os.path.dirname(pth),build_context)).resolve()
      # TODO make sure this is in the right repository
      #      If the build context of the stage goes up above the repo directory this will be wrong
    except Exception as exc:
      logging.error('Stage build context "{p}" for stage {n} may not be a directory'.format(p=build_context,n=name))
      return False

    item = {'name':name,
      'thingtype':'stage',
      'path':pth,
      'repo':reponame,
      'args':args,
      'secrets':secrets,
      'ssh':ssh,
      'build_context':str(build_context),
      'depends_on':dependency
    }
    return utils._add_item(item, items, VALIDATE)

TAGSAFE = lambda v : re.sub(r'[^a-zA-Z0-9\-\_]','_',v)
ARGTAG = lambda k,a : TAGSAFE(k) + '.' + TAGSAFE(a[k])
def taglist(_stage_name_, items):
  stage = items[_stage_name_]
  my_dep_args = [sa['key'] for sa in stage['args']]
  for s in stage['depends_on']:
    my_dep_args += [a for a in taglist(s, items) if a not in my_dep_args]
  return my_dep_args

def stage_args(stage_name, items, build_arg_dict=None):
  if build_arg_dict is None: build_arg_dict = {}
  my_dep_args = taglist(stage_name, items)
  this_args = {}
  for key in my_dep_args:
    if key in build_arg_dict:
      this_args[key] = build_arg_dict[key]
    else:
      logging.error('Building {s} requires setting build argument "{b}"'.format(s=stage_name,b=key))
      return False
  return this_args

def stage_prefix(stage_name):
  return 'vy__' + stage_name.replace('stage:','',1).lower()

def stages(stage_item):
  cmd = ['docker','images','--filter=reference=' + stage_prefix(stage_item) + ':*','--format','{{.Repository}}:{{.Tag}}']
  return subprocess.check_output(cmd).decode('utf-8').strip().split()

def stage_tag(stage_name, items, build_arg_dict=None):
  this_args = stage_args(stage_name, items, build_arg_dict)
  if this_args == False: return (False,False)
  keys = list(this_args.keys())
  keys.sort()
  lst = [ARGTAG(key, this_args) for key in keys]
  tag = stage_prefix(stage_name) + (':' + '.'.join(lst) if len(lst) > 0 else ':latest')
  return (tag, this_args)

def get_built_stage_repos(tag):
  cmd = ['docker', 'inspect', '--format', '\'{{ index .Config.Labels "tools.vy.repos"}}\'', tag]
  try:
    vyreposi = subprocess.check_output(cmd).decode('utf-8').strip().strip("'")
    return [] if len(vyreposi) == 0 else vyreposi.split(',')
  except Exception as exc:
    vytools.printer.print_def(' '.join(cmd)+'\n'+repr(exc))
  return []

def get_built_stage_versions(stage, items, build_args, checked):
  def _get_built_stage_versions(type_name, checked, top):
    if type_name not in items or not type_name.startswith('stage:'):
      return []
    item = items[type_name]
    tag, this_args = stage_tag(type_name, items, build_arg_dict=build_args)
    if tag and tag in checked:
      return checked[tag].copy()
    elif tag:
      stage_versions = []
      for d in item['depends_on']: 
        stage_versions += [sv for sv in _get_built_stage_versions(d, checked, False) if sv not in stage_versions]
      if not top:
        stage_versions += [sv for sv in get_built_stage_repos(tag) if sv not in stage_versions]
      elif top and item['repo'] in items:
        this_stage_repostr = utils.repo_version_string(items[item['repo']])
        if this_stage_repostr not in stage_versions:
          stage_versions.append(this_stage_repostr)
      checked[tag] = stage_versions.copy()
      return stage_versions

  stage_versions = _get_built_stage_versions('stage:' + stage['name'], checked, True)
  return sorted(stage_versions)

def __build__(stage, already_tagged, build_arg_dict, items, checked_versions, jobpath=None):
  if jobpath is None: jobpath = CONFIG.job_path()
  if not jobpath:
    return False, []
  ok,secret_cmds = CONFIG.secrets_cmd(stage['secrets'],stage['ssh'])
  if not ok:
    logging.error('Failed to build '+stage['name'])
    return False, []

  stage_name = 'stage:'+stage['name']
  dfile_dir = os.path.join(jobpath,'__dockerfiles__')
  os.makedirs(dfile_dir, exist_ok=True)
  tag = already_tagged[stage_name]['tag']

  contents = Path(stage['path']).read_text()
  text = '# syntax=docker/dockerfile:experimental\n'
  for (k,v) in build_arg_dict.items():
    text += 'ARG {k}="{v}"\n'.format(k=k,v=v)
  dep_stage_ids = []
  for s in stage['depends_on']:
    st = already_tagged[s]
    dep_stage_ids.append(st['tag']+'='+DOCKER_CLIENT.images.get(st['tag']).attrs['Id'])
    text += 'FROM {t} as {s}\n'.format(t=st['tag'], s=st['name'])
  for line in contents.splitlines(): #TODO nice to preserve backslash continuation
    text += line+'\n'
    if re.search(r'^[\s]*FROM ',line,re.I):
      text += 'LABEL tools.vy.remove=true\nLABEL tools.vy.tag=-\n'

  stage_versions = get_built_stage_versions(stage, items, build_arg_dict, checked_versions)
  text += 'LABEL tools.vy.repos={}\n'.format(','.join(stage_versions))
  text += 'LABEL tools.vy.images="{}"\n'.format(','.join(dep_stage_ids))
  text += 'LABEL tools.vy.tag={}\n'.format(tag)
  text += '# ----------------------------------------\n\n\n'

  build_context = stage['build_context']
  df = os.path.join(dfile_dir, tag+'.dockerfile')
  with open(df,'w') as w: w.write(text)
  cmd = ['docker', 'build', '--force-rm', '-f', df, '-t', tag]  # don't tag because of cleanup , '-t', node['s['tag']']
  cmd += ['--progress=plain']
  for (k,v) in build_arg_dict.items():
    cmd += ['--build-arg',k+'='+v]
  cmd += secret_cmds + [build_context]
  
  # vytools.printer.print_info(' '.join(cmd)+'\n\n'+text)
  with open(df+'.log','w') as logfile:
    vytools.printer.print_def(' '.join(cmd))
    proc = subprocess.Popen(cmd, env={'DOCKER_BUILDKIT':'1'}, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True)
    for line in proc.stdout:
      vytools.printer.print_def(line.strip(),fp=logfile)
    proc.wait()

  success = proc.returncode == 0
  if success:
    already_tagged[stage_name]['hash'] = subprocess.check_output(['docker','inspect','--format','{{.ID}}',tag]).decode('utf-8').strip()
    clean_up(['dangling=true','label=tools.vy.tag='+tag])
    # os.remove(df)
  else:
    logging.error('Failed to build '+tag)
  return success, stage_versions

def sorted_vy_images(filters):
  cmd = ['docker','images']
  for f in filters:
    cmd += ['--filter',f]
  cmd += ['--format','{{.CreatedAt}}\t{{.ID}}\t{{.Repository}}\t{{.Tag}}']
  imlist = subprocess.check_output(cmd).decode('ascii').strip()
  if len(imlist) > 0:
    images = []
    for im in imlist.split('\n'):
      imageid = im.split('\t')
      images.append({'created':imageid[0], 'id':imageid[1], 'repo':imageid[2], 'tag':imageid[3]})
    return sorted(images, key=lambda k: k['created'], reverse=True) # Most recent first
  return []

def clean_up(filters):
  subprocess.run(['docker','builder','prune','-f']) # Todo, why do I need this? Why is their a build cache when everything builds successfully?
  untagged = [im['id'] for im in sorted_vy_images(filters)]
  untagged.reverse()
  if len(untagged) > 0:
    while True:
      try:
        logging.info('Attempting to remove dangling images:'+str(untagged))
        subprocess.check_output(['docker','rmi']+untagged)
        break
      except Exception as exc:
        logging.warning(str(exc))
        time.sleep(5)

def find_all(items, contextpaths=None):
  # r'[.]*dockerfile[.]*'
  return utils.search_all(r'(.+)\.stage\.dockerfile', parse, items, contextpaths=contextpaths)

def build(stages, items, anchors, already_built, build_level, jobpath=None):
  # build_level = 1 # Rebuild all - unless you've already
  # build_level = 0 # Rebuild top level - unless you've already sometime 
  # build_level = -1 # Don't build anything
  if not utils.exists(stages, items):
    logging.error('  Failed to build {}'.format(stages))
    return False

  existing_images = [im['repo']+':'+im['tag'] for im in sorted_vy_images(['label=tools.vy.tag'])]
  checked_versions = {}
  already_tagged = {}
  def __multistage__(stage_name):
    stage = items[stage_name]
    for s in stage['depends_on']:
      if __multistage__(s) == False: return False
    if stage_name not in already_tagged:
      tag,this_args = stage_tag(stage_name, items, anchors)
      if not tag: return False
      already_tagged[stage_name] = {'tag':tag,'name':stage['name'],'skipped':False,'stage_versions':[]}
      oktobuild = ((build_level == 0 and stage_name in stages) or
                  (build_level == 1) or (tag not in existing_images)) and tag not in already_built
      if oktobuild:
        vytools.printer.print_info(' * Building: {s} --as-- {i}'.format(s=stage_name,i=tag))
        success, stage_versions = __build__(stage, already_tagged, this_args, items, checked_versions, jobpath) # Build this stage
        if not success:
          vytools.printer.print_fail(' * Failed {s} --as-- {i}'.format(s=stage_name,i=tag))
          return False
        if stage_versions:
          already_tagged[stage_name]['stage_versions'] = stage_versions
        already_built.append(tag)
        vytools.printer.print_success(' * Finished: {s} --as-- {i}\n\n'.format(s=stage_name,i=tag))
      elif tag not in already_built:
        already_tagged[stage_name]['stage_versions'] = get_built_stage_versions(stage, items, this_args, checked_versions)
      else:
        already_tagged[stage_name]['skipped'] = True
        # # I think this is cumbersome/notuseful printing
        # if tag not in already_built:
        #   vytools.printer.print_info('* Skipping: {s} as {i}'.format(s=stage_name,i=tag))
        # else: 
        #   vytools.printer.print_info('* Already built: {s} as {i}'.format(s=stage_name,i=tag))
    return True

  for top_name in stages:
    if False == __multistage__(top_name):
      return False

  if build_level != -1:
    for tag in already_tagged:
      if not already_tagged[tag]['skipped']:
        vytools.printer.print_success(' * Built {n} --as-- {t}'.format(n=tag,t=already_tagged[tag]['tag']))
      # else: # I think this is cumbersome/notuseful printing
        # vytools.printer.print_info(' * Skipped {n} as {t}'.format(n=tag,t=already_tagged[tag]['tag']))
  return already_tagged #sum([int(v['hash'],16) for v in already_tagged.values()])

def run(type_name, items=None, anchors=None, jobpath=None, cmd=None):
  if items is None: items = ITEMS
  if anchors is None: anchors = {}
  already_built = []
  built = build([type_name], items, anchors, already_built, -1, jobpath=jobpath)
  if built and type_name in built:
    if jobpath is None: jobpath = CONFIG.job_path()
    if not jobpath:
      return False
    tag = built[type_name]['tag']
    datadir = os.path.join(jobpath,'__data__')
    os.makedirs(datadir,exist_ok=True)
    vytools.printer.print_info('** Running {t} with "{p}" mounted to "/__vydata__" in the container'.format(t=tag,p=datadir))
    dcmd = ['docker','run','--rm', '-it', '-v', datadir+':/__vydata__', tag]
    if cmd is not None: 
      dcmd += shlex.split(cmd)
    proc = subprocess.run(dcmd,stderr=subprocess.STDOUT)
    # else:
    #   volumes = {}
    #   volumes[datadir] = {'bind':'/__vydata__','mode':'rw'}
    #   x=DOCKER_CLIENT.containers.run(tag,cmd,remove=True,volumes=volumes)
    #   print(x)



import vytools.utils as utils
import vytools.stage
from vytools.config import ITEMS
from vytools.composerun import run
import yaml, json, os, shutil, copy, re
import cerberus
import logging

KEYSRULES = '[a-zA-Z0-9_]+'
SCHEMA = utils.BASE_SCHEMA.copy()
SCHEMA.update({
  'thingtype':{'type':'string', 'allowed':['compose']},
  'ui':{'type':'string', 'maxlength': 64, 'required': False},
  'subcompose':{'type':'list',
    'schema': {
      'type': 'dict',
      'schema': {
        'name': {'type': 'string', 'maxlength': 64},
        'calibration':{'type': 'dict', 'required':False}
      }
    }
  },
  'anchors':{'type': 'dict', 'required': False, 'keysrules': {'type': 'string', 'regex': KEYSRULES}}
})
VALIDATE = cerberus.Validator(SCHEMA)

def parse(name, pth, reponame, items=None):
  if items is None: items = vytools.ITEMS
  item = {
    'name':name,
    'thingtype':'compose',
    'repo':reponame,
    'depends_on':[],
    'path':pth
  }
  itemx = {'ui':None, 'subcompose':[], 'anchors':{}}
  try:
    with open(pth,'r') as r:
      content = yaml.safe_load(r.read())
      xvy = content.get('x-vy',{})
      itemx.update((k, xvy[k]) for k in itemx.keys() & xvy.keys())
      if itemx['ui'] is None: del itemx['ui']
      item.update(itemx)
    return utils._add_item(item, items, VALIDATE)
  except Exception as exc:
    logging.error('Failed to parse/load compose file "{n}" at "{p}":\n    {e}'.format(n=name,p=pth,e=exc))
    return False
ANCHORTHINGTYPES = ['stage','definition','vydir']
ANCHORTYPES = ['definition','argument','artifact','stage','vydir','directory','file']

def find_all(items, contextpaths=None):
  success = utils.search_all(r'(.+)\.compose\.y[a]*ml', parse, items, contextpaths=contextpaths)
  if success: # process definitions
    for (type_name, item) in items.items():
      if type_name.startswith('compose:'):
        successi = True
        item['depends_on'] = []
        for e in item['anchors']:
          atype = [a for a in ANCHORTYPES if item['anchors'][e].startswith(a+':')]
          if len(atype) == 1 and atype[0] in ANCHORTHINGTYPES: 
            successi &= utils._check_add(item['anchors'][e], atype[0], item, items, type_name)
          elif len(atype) != 1:
            successi = False
            logging.error('Unknown anchor type "{t} {v}" in {n}. Should be one of: {a}'.format(t=e,v=item['anchors'][e],n=type_name,a=','.join(ANCHORTYPES)))
        if 'ui' in item:
          successi &= utils._check_add(item['ui'].split('/')[0], 'vydir', item, items, type_name)
        for e in item['subcompose']:
          successi &= utils._check_add(e['name'], 'compose', item, items, type_name)
        success &= successi
        if not successi:
          logging.error('Failed to interpret/link compose {c}'.format(c=type_name))
  return success

def build(rootcompose, items=None, anchors=None, built=None, build_level=1, object_mods=None, eppath=None, label=None):
  if items is None: items = ITEMS
  if anchors is None: anchors = {} # These are the anchor values specified by the episode or otherwise
  if built is None: built = []
  if label is None: label = rootcompose.replace('compose:','') 
  if not utils.exists([rootcompose], items):
    return False
  elif not rootcompose.startswith('compose:'):
    logging.error('Item {} is not a compose file'.format(rootcompose))
    return False

  stage_versions = []
  item = items[rootcompose]

  cmd = []
  dependencies = [rootcompose]

  child_anchors = {} # Populate with the merged/filled out anchor values from subcomposes
  for sa in item['subcompose']:
    subname = sa['name']
    sublabel = label + '.'+subname.replace('compose:','')
    subcmds = build(subname, items=items, anchors=copy.deepcopy(anchors), built=built,
                            build_level=build_level, object_mods=object_mods, eppath=eppath, label=sublabel)
    if subcmds == False:
      return False
    else:
      child_anchors.update(subcmds['anchors']) # Brutal overwrites not namespacing. I don't think I want namespacing
      cmd += subcmds['command']
      dependencies += [d for d in subcmds['dependencies'] if d not in dependencies]
      stage_versions += [sv for sv in subcmds['stage_versions'] if sv not in stage_versions]

  anchor_defaults = copy.deepcopy(item.get('anchors',{}))
  anchor_values = copy.deepcopy(anchors)
  ordered = []
  for atype in ANCHORTYPES:
    ordered += [k for k,v in anchor_defaults.items() if v.startswith(atype+':')]
  replace_obj = {}
  for tag in ordered: # SORTED TO ENSURE THIS ORDER DEFINITION, ARGUMENT, STAGE, ARTIFACT
    val = anchor_defaults[tag]
    if val.startswith('definition:'):
      obj_mods = object_mods[tag] if object_mods and tag in object_mods else None
      if tag not in anchor_values:
        logging.warning('Anchor "{}" was not set for use in compose "{}"'.format(tag,rootcompose))
        replace_obj[tag] = {}
        deplist_ = []
      else:
        replace_obj[tag],deplist_ = vytools.object.expand(anchor_values[tag],
            val, items, object_mods=obj_mods)
      if replace_obj[tag] is None:
        return False
      dependencies += [d for d in deplist_ if d not in dependencies]
    elif val.startswith('argument:'):
      replace_obj[tag] = val.replace('argument:','',1)
      if tag not in anchor_values: 
        anchor_values[tag] = replace_obj[tag]
    elif val.startswith('stage:'):
      buildstage = anchor_values[tag] if tag in anchor_values else val
      tagged = vytools.stage.build([buildstage], items, anchor_values, built, build_level, jobpath=eppath)
      if tagged == False: return False
      for k,v in tagged.items():
        stage_versions += [sv for sv in v['stage_versions'] if sv not in stage_versions]
      replace_obj[tag] = tagged[buildstage]['tag']
    elif any([val.startswith(v+':') for v in ['directory','file','artifact','vydir']]):
      val_ = anchor_values[tag] if tag in anchor_values else val
      splitname = val_.split(':',1)
      arttyp = splitname[0]
      artname = splitname[-1]
      if len(splitname) == 2 and len(artname) > 0:
        if arttyp == 'vydir':
          artname = vytools.utils.get_vd_path(val_, items)
          if not artname:
            return False
        replace_obj[tag] = artname
        if build_level == -1 and '..' not in artname and eppath and os.path.exists(eppath):
          artifactpath = os.path.join(eppath, artname)
          if arttyp in ['artifact','file']:
            with open(artifactpath,'w') as w: w.write('')
            os.chmod(artifactpath, 0o666)
          elif arttyp == 'directory':
            os.makedirs(artifactpath,exist_ok=True)

  if build_level == -1:
    compose_file_name = label + '.yaml'
    compose_pth = item['path']
    composition = {}
    with open(compose_pth,'r') as r:
      composition = yaml.safe_load(r.read())
    if 'x-vy' in composition: del composition['x-vy']
    _recursiv_replace(composition, replace_obj, eppath)
    child_anchors.update(replace_obj)
    
    oktowrite = '..' not in compose_file_name and eppath and os.path.exists(eppath)
    if oktowrite:
      cfile = os.path.join(eppath, compose_file_name)
      if not bool(replace_obj):
        shutil.copyfile(compose_pth, cfile)
      else:
        with open(cfile,'w') as w:
          w.write(yaml.safe_dump(composition))
      cmd = cmd + ['-f',compose_file_name]
  
  return {
    'anchors':child_anchors,
    'command':cmd,
    'stage_versions':sorted(stage_versions),
    'dependencies':dependencies
  }

def _prefx(key,char,replkeys):
  return [r for r in replkeys if key.startswith(r+char)]

def stripkey(key):
  if key.startswith('$'):
    key = key[1:]
    if key.startswith('{'):
      key = ''.join(key[1:].split('}',1))
  return key

def _replkeysf(key, repl, eppath):
  replkeys = [str(i) for i in len(repl)] if type(repl) == list else \
    (repl.keys() if type(repl) == dict else [])
  if key in replkeys:
    return repl[key]
  else:
    if type(repl) == dict:
      for kkey in replkeys:
        if key.startswith(kkey+':') or key.startswith(kkey+'/'):
          return key.replace(kkey,repl[kkey],1)
  for x in ['.','>']:
    prefx = _prefx(key,x,replkeys)
    if len(prefx) == 1:
      repl_ = repl[prefx[0]]
      key_ = stripkey(key.replace(prefx[0]+x,'',1))
      if x == '.':
        return _replkeysf(key_, repl_, eppath)
      elif x == '>' and eppath and os.path.exists(eppath) and '..' not in key_:
        path = os.path.join(eppath, key_)
        with open(path,'w') as w:
          if key.endswith('.json'):
            w.write(json.dumps(repl_))
          elif key.endswith('.yaml'):
            w.write(yaml.safe_dump(repl_))
        return path

def _recursiv_replace(obj, repl, eppath):
  if type(obj) == dict:
    ks = obj.keys()
  elif type(obj) == list:
    ks = range(len(obj))
  else:
    return
  for k in ks:
    if type(obj[k]) == str:
      # Find environment variables e.g. ${a} $a  $a.b.c  ${a.b.c}
      keys = re.findall(r'\$[\{]?[a-zA-Z0-9_\.\>]+[\}]?',obj[k],re.I)
      for key in keys:
        val = _replkeysf(key.strip('$').strip('{').strip('}'), repl, eppath)
        if val: obj[k] = obj[k].replace(key, str(val))
    elif type(obj[k]) in [list,dict]:
      _recursiv_replace(obj[k], repl, eppath)
      
def artifact_paths(compose_name, items, eppath):
  artifacts = {}
  def get_artifacts(i,artifacts):
    if i in items:
      for tag,val in items[i].get('anchors',{}).items():
        pth = os.path.join(eppath,val.replace('artifact:','',1))
        if val.startswith('artifact:') and '..' not in val and os.path.exists(pth):
          artifacts[tag] = pth
      for sa in items[i]['subcompose']: get_artifacts(sa['name'],artifacts)
  if eppath and os.path.exists(eppath):
    get_artifacts(compose_name,artifacts)
  return artifacts



import os, re, logging, subprocess, re
from vytools.config import CONFIG, SEARCHED_REPO_PATHS
import vytools.printer

BASE_DATA_TYPES = {'float32':0,'float64':0,'uint64':0,
  'int64':0,'uint32':0,'int32':0,'uint16':0,'int16':0,'uint8':0,
  'int8':0,'string':'','bool':False,'byte':0,'char':''}

BASE_SCHEMA = {
  'path': {'type':'string', 'maxlength': 1024},
  'name': {'type':'string', 'maxlength': 64},
  'depends_on':{'type':'list','schema': {'type': 'string', 'maxlength':64}},
  'repo':{'type':'string', 'maxlength': 64}
}

REPO_SCHEMA = BASE_SCHEMA.copy()
REPO_SCHEMA.update({
  'source': {'type':'string', 'maxlength': 32},
  'account': {'type':'string', 'maxlength': 32},
  'reponame': {'type':'string', 'maxlength': 128},
  'url': {'type':'string', 'maxlength': 512},
  'type': {'type':'string', 'allowed': ['git','hg']},
})

def get_vd_path(name, items):
  if '..' in name or not name.startswith('vydir:'): return None
  ns = name.split('/',1)
  pth = items.get(ns[0],{}).get('path',None)
  if not pth:
    logging.error('Failed to find a vy directory "{}"'.format(ns[0]))
    return None
  pth = os.path.join(pth,ns[1]) if len(ns)==2 else pth
  if not os.path.exists(pth):
    logging.error('The directory/file at "{}" does not exist'.format(pth))
    return None
  return pth
  
def find_all_vydirectories(items, contextpaths=None):
  success = search_all(None, None, items, checkvydir=True, contextpaths=contextpaths)
  return success

def repo_version(typ, pth):
  try:
    if typ == 'hg':
      return subprocess.check_output(['hg','--cwd',pth,'id','-i']).decode('ascii').strip()  # unfortunately THIS CHANGES FILES IN THE .Hg folder
    elif typ == 'git':
      hash_ = subprocess.check_output(['git','-C',pth,'rev-parse','HEAD']).decode('ascii').strip()
      changes = subprocess.check_output(['git','-C',pth,'status','--porcelain']).decode('ascii').strip()
      hash_ += '+' if len(changes) > 0 else ''
      return hash_ #subprocess.check_output(['git','-C',pth,'rev-parse','HEAD']).decode('ascii').strip()
    else:
      return '-'
  except Exception as exc:
    logging.error('Failed to get hash of repo: '+str(exc))
    return '-'

def parse_repo(pth, items):
  if pth in SEARCHED_REPO_PATHS: return SEARCHED_REPO_PATHS[pth]
  SEARCHED_REPO_PATHS[pth] = False

  if os.path.exists(os.path.join(pth,'.git')): # Can be directory or file so use exists
    typ = 'git'
    options = ['origin','remote'] # to do more here
    cmdprefx = ['git','-C',pth,'config','--get']
  elif os.path.isdir(os.path.join(pth,'.hg')):
    typ = 'hg'
    options = ['default'] # to do more here
    cmdprefx = ['hg','--cwd',pth,'paths']
  else:
    return SEARCHED_REPO_PATHS[pth]

  s = {'type':typ,'path':pth,'depends_on':[],'version':'','thingtype':'repo'}
  url = None
  for o in options: 
    try:
      cmd = cmdprefx + ['remote.{o}.url'.format(o=o) if typ == 'git' else o]
      url = subprocess.check_output(cmd).decode('utf-8').strip()
      if len(url) > 0: break
    except Exception as exc:
      pass
  if not url:
    logging.error('Failed to get remote for repository at {p}'.format(p=pth))
    return SEARCHED_REPO_PATHS[pth]

  # e.g.
  # git@bitbucket.org:account/name.git
  # https://username@bitbucket.org/account/name.git
  # https://github.com/account/name.git
  # git@github.com:account/name.git
  s['url'] = url
  urlparts = url.split('@',1)[-1]
  urlparts = re.split('/|:',urlparts)
  if len(urlparts) < 3:
    logging.error('Failed to parse url "{u}" for repository at {p}'.format(u=url,p=pth))
    return SEARCHED_REPO_PATHS[pth]

  s['source'] = urlparts[0]
  s['account'] = urlparts[1]
  s['reponame'] = urlparts[2].replace('.git','').replace('.hg','')
  name_root = s['account']+'|'+s['reponame']
  name = name_root
  count = 0
  while 'repo:'+name in items:
    name = name_root+'|'+str(count)
    count += 1
  s['repo'] = 'repo:'+name
  s['name'] = name
  items['repo:'+name] = s
  SEARCHED_REPO_PATHS[pth] = True
  return SEARCHED_REPO_PATHS[pth]

def repo_version_string(repo, version=None):
  if version == None:
    version = repo_version(repo['type'], repo['path'])
  return repo['type'] + '|' + repo['source'] + '|' + repo['name'] + '|' + version

def get_repo_versions(lst, items, build_args=None):
  repo_versions = {}
  def get_version(name):
    if name in items:
      item = items[name]
      if len(item['repo']) > 0 and item['repo'] in items and item['repo'] not in repo_versions:
        repo_versions[item['repo']] = repo_version_string(items[item['repo']])
      for d in item['depends_on']: get_version(d)
  for l in lst: get_version(l)
  return repo_versions

def get_repository_paths(rep_str_lst, contextpaths=None):
  sp = contextpaths if contextpaths else CONFIG.get('contexts')
  items = {}
  newcp = []
  success = True
  search_all(None, None, items, contextpaths=sp, find_repos_only=True)
  repo_versions = get_repo_versions([i for i in items],items)
  vytools.printer.print_def('Found repositories:')
  for item in sorted([v for v in repo_versions.values()]):
    vytools.printer.print_def('  '+item)
  for r in rep_str_lst:
    found = False
    for i,item in repo_versions.items():
      found = item.startswith(r)
      if found:
        newcp.append(items[i]['path'])
        break
    if not found:
      success = False
      logging.error('Failed to find repository matching {n} in the searched directories {p}'.format(n=r,p=','.join(sp)))
  SEARCHED_REPO_PATHS.clear()
  return (success, newcp)

def is_vydir(files):
  vydir = None
  for f in files:
    if f.endswith('.vydirectory'):
      vydir = f.replace('.vydirectory','',-1)
      break
    if f.endswith('.vydir'):
      vydir = f.replace('.vydir','',-1)
      break
  if vydir and vydir == 'vy':
    logging.error('"vy" is a reserved vydirectory name')
    vydir = None
  return vydir

def search_all(fname_regex, func, items, checkvydir=False, contextpaths=None, find_repos_only=False):
  if contextpaths is None: contextpaths = CONFIG.get('contexts')
  success = False
  if contextpaths:
    success = True
    exclude = set(['.vy','.git','.hg'])
    for cp in contextpaths:
      for root, dirs, files in os.walk(cp, topdown=True):
        if '.vyignore' in files:
          dirs[:] = []
          continue
        isrepo = parse_repo(root, items)
        dirs[:] = [d for d in dirs if d not in exclude]
        if isrepo and find_repos_only:
          dirs[:] = []
          continue
        if checkvydir:
          vydir = is_vydir(files)
          if vydir: 
            item = {'name':vydir, 'repo':'', 'thingtype':'vydir', 'depends_on':[], 'path':root}
            success &= _add_item(item, items, True)
        elif fname_regex and func:
          for f in files:
            m = re.match(fname_regex,f,re.I)
            if m:
              success &= func(m.group(1), os.path.join(root, f), '', items)
  return success

def topological_sort(source):
    pending = [(name, set(deps)) for name, deps in source]        
    emitted = []
    while pending:
      next_pending = []
      next_emitted = []
      for entry in pending:
        name, deps = entry
        deps.difference_update(set((name,)), emitted)
        if deps:
          next_pending.append(entry)
        else:
          yield name
          emitted.append(name)
          next_emitted.append(name)
      if not next_emitted:
        raise ValueError("cyclic dependancy detected {n}".format(n=name))
      pending = next_pending
      emitted = next_emitted
    return emitted

def exists(lst, items, pad=''):
  success = True
  for l in lst:
    if l not in items:
      success = False
      logging.error('"{n}" was not found {p}'.format(n=l,p=pad))
    else:
      success &= exists(items[l]['depends_on'],items,'(depended on by {})'.format(l))
  return success

def sort(lst, items):
  return [x for x in topological_sort([(k,set(items[k]['depends_on'])) for k in lst])]

def _check_add(nme, typ, item, items, parent):
  if nme.startswith(typ+':') and nme in items:
    item['depends_on'].append(nme)
    return True
  elif typ == 'vydir' and nme.startswith(typ+':') and nme.split('/')[0] in items:
    item['depends_on'].append(nme.split('/')[0])
    return True
  logging.error('Failed to find subitem "{n}" referenced by "{p}".\n  - "{n}" is not of type {t} or does not exist'.format(n=nme, t=typ, p=parent))
  return False

def _add_item(item, items, validate):
  typ = item['thingtype']
  name = item['name']
  pth = item['path']
  tname = typ+':'+name
  if tname in items:
    pthp = items[tname]['path']
    logging.error('"{t}" at "{p}" was not loaded because a same name item was already loaded from {p2}' \
                  .format(t=tname, p=pth, p2=pthp))
    return False
  elif validate==True or validate.validate(item):
    items[tname] = item
    return True
  else:
    logging.error('"{n}" at "{p}" failed validation {s}'.format(n=tname, p=pth, s=validate.errors))
    return False

def missing_item(type_name, items):
  if not items:
    logging.error('No items are included. Do you have items defined? Have you scanned for them?')
  else:
    examples = []
    for k in items.keys():
      if not any([e.startswith(k.split(':')[0]+':') for e in examples]):
       examples.append(k)
    logging.error('Item "{}" not found in vy items. Check the spelling and format.  Example items:\n  {}'.format(type_name,'\n  '.join(examples)))



#!/usr/bin/env python3
import argparse, sys, os, json, argcomplete, re
import subprocess
import vytools

import logging
logging.basicConfig(level=logging.INFO)

TLIST = '(repo, vydir, definition, object, stage, compose, episode)'

def add_all(parser):
  parser.add_argument('--all', action='store_true', 
                      help='If present, rebuild all dependent stages. '
                            'Otherwise only rebuild top level and missing stages')

def make_parser():
  parser = argparse.ArgumentParser(prog='vytools', description='tools for working with vy')
  parser.add_argument('--version','-v', action='store_true', 
    help='Print vytools version')
  parser.add_argument('--contexts', type=str, default='',
    help='Comma or semi-colon delimited list of paths to context folders')
  parser.add_argument('--secret', metavar='KEY=VALUE', action = 'append', required=False,  default=[],
    help='Key-value pair for docker build secrets e.g. --secret MYSECRET=/path/to/my_secret.txt '
                            "(do not put spaces before or after the = sign). "
                            "Value is the path to the file containing the secret. If the path "
                            "contains spaces, you should define it with double quotes: "
                            'foo="/path/to some/place/secret".')
  parser.add_argument('--ssh', metavar='KEY=VALUE', action = 'append', required=False, default=[],
    help='Key-value pair for ssh keys e.g. --ssh MYKEY=/path/to/my_key '
                            "(do not put spaces before or after the = sign). "
                            "Value is the path to the ssh key. If the path "
                            "contains spaces, you should define it with double quotes: "
                            'foo="/path/to some/place/sshkey".')      
  parser.add_argument('--jobs', type=str, required=False, 
    help='Path to jobs folder. All jobs will be written to this folder.')
  parser.add_argument('--anchor','-a', metavar='KEY=VALUE', action = 'append', required=False,
                      help="Set key-value pairs "
                            "(do not put spaces before or after the = sign). "
                            "If a value contains spaces, you should define "
                            "it with double quotes: "
                            'foo="this is a sentence". Note that '
                            "values are always treated as strings.")      
  parser.add_argument('--name','-n', action='append', choices=vytools.CONFIG.get('items'),
    help='Name of thing '+TLIST+' ', metavar='')

  subparsers = parser.add_subparsers(help='specify action', dest='action')

  build_sub_parser = subparsers.add_parser('build',
    help='Build docker images that are dependent on named items')
  add_all(build_sub_parser)

  grep_sub_parser = subparsers.add_parser('grep',
    help='Grep logs or things '+TLIST)
  grep_sub_parser.add_argument('--pattern','-p', type=str, required=True, help='grep pattern')
  grep_sub_parser.add_argument('--flags','-f', action='append', help='grep flags, use a space before a dash. e.g. vytools grep --pattern "error:" -flags " -i" --lastlog', metavar='')
  grep_sub_parser.add_argument('--lastlog','-l', action='store_true', help='grep the last log')

  info_sub_parser = subparsers.add_parser('info',
    help='Print things '+TLIST)
  info_sub_parser.add_argument('--images','-i', action='store_true', 
    help='List images')
  info_sub_parser.add_argument('--private','-p', action='store_true', 
    help='List ssh/secrets')
  info_sub_parser.add_argument('--dependencies','-d', action='store_true', 
    help='List dependencies of items')
  info_sub_parser.add_argument('--expand','-e', action='store_true', 
    help='Expand items')

  run_sub_parser = subparsers.add_parser('run',
    help='Run specified stage/compose/episode items')
  run_sub_parser.add_argument('--save','-s', action='store_true', 
    help='If present, save episodes.')
  run_sub_parser.add_argument('--build', action='store_true', 
    help='If present, build dependent stages (also note --all flag).')
  add_all(run_sub_parser)
  run_sub_parser.add_argument('--persist','-p', action='store_true', 
    help='Persist created docker volumes after running')
  run_sub_parser.add_argument('--clean', action='store_true', 
    help='Clean episode folders before running')
  run_sub_parser.add_argument('--object-mods','-o', metavar='KEY=VALUE', action = 'append', required=False,
                      help='Set key-value pairs of objects '
                            '(do not put spaces before or after the = sign). '
                            'If a value contains spaces, you should define '
                            'it with double quotes: foo="this is a sentence". Note that '
                            'values are always treated as strings.')  
  run_sub_parser.add_argument('--cmd', type=str, required=False, 
    help='Stage command')

  return parser

def parse_key_value(kv,typ):
  args = {}
  success = True
  if kv:
    for arg in kv:
      if '=' not in arg:
        success = False
        logging.error('A {s} ({a}) failed to be in the form KEY=VALUE'.format(s=typ,a=arg))
      else:
        (k,v) = arg.split('=',1)
        args[k] = v
  return (success, args)

def parse_anchors(ba):
  return parse_key_value(ba,'anchor (--anchor, -a)')

def parse_object_mods(args):
  oa = args.object_mods if 'object_mods' in dir(args) and args.object_mods else []
  (success, obj) = parse_key_value(oa,'object-mods (--object-mods, -o)')
  if success:
    for k,v in obj.items():
      try:
        obj[k] = json.loads(v)
      except Exception as exc:
        logging.error('Failed to parse input object-mods:'+str(exc))
        success = False
  return (success, obj)

def grep(lst,pattern,flags,lastlog):
  cmd = ['grep',pattern]
  out = ''
  if flags: cmd += [f.strip() for f in flags]
  if lastlog:
    path = os.path.join(vytools.CONFIG.job_path(),'__dockerfiles__')
    paths = [os.path.join(path, basename) for basename in os.listdir(path) if basename.endswith('.log')]
    if len(paths) > 0:
      out = subprocess.check_output(cmd+[max(paths, key=os.path.getctime)]).decode('ascii')
  else:
    pass
  print(out)

def main():
  notScanningPreexistingContext = vytools.CONFIG.get('contexts') and '--contexts' not in sys.argv
  notAskingForHelp = '-h' not in sys.argv and '--help' not in sys.argv
  notAskingForVersion = '-v' not in sys.argv and '--version' not in sys.argv
  someArguments = len(sys.argv)>1
  if notScanningPreexistingContext and someArguments and notAskingForHelp and notAskingForVersion:
    vytools.printer.print_def('scanning...')
    vytools.scan()
  parser = make_parser()
  argcomplete.autocomplete(parser)
  args = parser.parse_args()
  if args.version:
    vytools.printer.print_def(vytools.__version__)
    return

  if args.contexts: 
    vytools.CONFIG.set('contexts', re.split(';|,',args.contexts))
    vytools.scan()
  if args.jobs: vytools.CONFIG.set('jobs',args.jobs)
  if not vytools.CONFIG.get('contexts'):
    logging.error('The vy context(s) has not been initialized. Please provide a --contexts')
    return

  if 'ssh' in dir(args) and args.ssh:
    success, sshkeys = parse_key_value(args.ssh,'argument (--ssh)')
    if not success: return
    vytools.CONFIG.set('ssh', sshkeys)
  if 'secret' in dir(args) and args.secret:
    success, secrets = parse_key_value(args.secret,'argument (--secret)')
    if not success: return
    vytools.CONFIG.set('secrets', secrets)

  lst = [n for n in args.name] if 'name' in dir(args) and args.name else []
  anchor_args = None if 'anchor' not in dir(args) else args.anchor
  success, anchors = parse_anchors(anchor_args)
  if not success: return
  build_level = 1 if 'all' in dir(args) and args.all else 0
  persist = args.persist if 'persist' in dir(args) else False

  if args.action == 'build':
    return bool(vytools.build(lst, anchors=anchors, build_level=build_level))
  elif args.action == 'run':
    success, object_mods = parse_object_mods(args)
    if not success: return
    cmd = None if 'cmd' not in dir(args) else args.cmd
    if object_mods == False:
      return False
    if args.build:
      br = vytools.build(lst, anchors=anchors, build_level=build_level)
      if br == False: return False
    return bool(vytools.run(lst, anchors=anchors, clean=args.clean, save=args.save, object_mods=object_mods, cmd=cmd, persist=persist))
  elif args.action == 'info':
    vytools.info(lst, list_dependencies=args.dependencies, list_private=args.private, list_images=args.images, expand=args.expand)
  elif args.action == 'grep':
    flags = args.flags if 'flags' in dir(args) else None
    lastlog = args.lastlog if 'lastlog' in dir(args) else None
    grep(lst, args.pattern, flags, lastlog)
  else:
    return False




import vytools.utils as utils
import yaml, re, json, io
import cerberus
import logging

SCHEMA = utils.BASE_SCHEMA.copy()
SCHEMA.update({
  'thingtype':{'type':'string', 'allowed':['definition']},
  'element':{
    'type':'list',
    'schema': {
      'type': 'dict',
      'schema': {
        'name': {'type': 'string', 'maxlength': 64},
        'optional': {'type': 'boolean', 'required':False},
        'length': {'type': 'string'},
        'type': {'type': 'string'}
      }
    }
  }
})
VALIDATE = cerberus.Validator(SCHEMA)

def parse(name, pth, reponame, items):
  item = {
    'name':name,
    'repo':reponame,
    'thingtype':'definition',
    'depends_on':[],
    'element':[],
    'path':pth
  }
  try:
    content = json.load(io.open(pth, 'r', encoding='utf-8-sig'))
    item['element'] = content['element']
  except Exception as exc:
    logging.error('Failed to parse definition "{n}" at "{p}": {e}'.format(n=name, p=pth, e=exc))
    return False

  return utils._add_item(item, items, VALIDATE)

def find_all(items, contextpaths=None):
  success = utils.search_all(r'(.+)\.definition\.json', parse, items, contextpaths=contextpaths)
  if success: # process definitions
    for (type_name, item) in items.items():
      if type_name.startswith('definition:'):
        (typ, name) = type_name.split(':',1)
        item['depends_on'] = []
        for e in item['element']:
          if e['type'] in utils.BASE_DATA_TYPES:
            pass
          elif e['type'] in items:
            item['depends_on'].append(e['type'])
          else:
            success = False
            logging.error('definition "{n}" has an invalid element type {t}'.format(n=name, t=e['type']))

          
  return success



import logging, json, copy
from pathlib import Path
from vytools.config import CONFIG, ITEMS

import vytools.utils
import vytools.stage
import vytools.compose
import vytools.episode
import vytools.definition
import vytools.object
import vytools.printer

def _label_repositories(items):
  repo_path_list = []
  repo_path_dict = {}
  for r,v in items.items():
    if r.startswith('repo:'):
      repo_path_list.append(v['path'])
      repo_path_dict[v['path']] = r
  repo_path_list.sort(key=len)
  repo_path_list.reverse()
  for i,v in items.items():
    ipaths = [v['path']]+[str(p) for p in Path(v['path']).parents]
    for r in repo_path_list:
      if r in ipaths:
        v['repo'] = repo_path_dict[r]
        break

def scan(contextpaths=None):
  if contextpaths:
    vytools.printer.print_def('\nScanning the following directories for vy items:'+'\n  '.join(['']+contextpaths)+'\n\n')
  CONFIG.set('scanned',contextpaths)

  items = {}
  success = True
  success &= vytools.definition.find_all(items, contextpaths=contextpaths)
  success &= vytools.object.find_all(items, contextpaths=contextpaths)
  success &= vytools.stage.find_all(items, contextpaths=contextpaths)
  success &= vytools.utils.find_all_vydirectories(items, contextpaths=contextpaths)
  success &= vytools.compose.find_all(items, contextpaths=contextpaths)
  success &= vytools.episode.find_all(items, contextpaths=contextpaths)

  # Assign repositories
  _label_repositories(items)

  # if success: # Check for missing dependencies
  for i,it in items.items():
    for dependency in it['depends_on']:
      if dependency not in items:
        logging.warning(' * Item {j} (depended on by {i}) is not in the list of found items'.format(i=i, j=dependency))
        success = False

  # if success: # Reverse dependency chain
  for i in items:
    items[i]['depended_on'] = []
  for i in items:
    for d in items[i]['depends_on']:
      if d in items:
        items[d]['depended_on'].append(i)

  if success:
    thinglist = [item for item in items]
    sorted_things = vytools.utils.sort(thinglist, items)
    for itemid in items: # Sort the depends_on of each thing
      items[itemid]['depends_on'] = [i for i in sorted_things if i in items[itemid]['depends_on']]

  ITEMS.clear()
  ITEMS.update(items)
  CONFIG.set('items',[i for i in items]) # Always write items?
  return success

def build(original_list, items=None, anchors=None, build_level=0, jobpath=None):
  if items is None: items = ITEMS
  if anchors is None: anchors = {}
  built = []
  stages = []
  for type_name in original_list:
    if type_name.startswith('stage:') and type_name not in stages:
      stages.append(type_name)
    elif type_name.startswith('compose:') and type_name in items:
      anchor_copy = copy.deepcopy(anchors)
      build_level = max(0,build_level) # Don't write for a compose if run at this level
      if False == vytools.compose.build(type_name, items=items, anchors=anchor_copy, built=built, build_level=build_level, eppath=jobpath):
        return False
    elif type_name.startswith('episode:') and type_name in items:
      anchor_copy = copy.deepcopy(anchors)
      build_level = max(0,build_level) # Don't write for a compose if run at this level
      if False == vytools.episode.build(type_name, built, anchors=anchor_copy, items=items, build_level=build_level, eppath=jobpath):
        return False

  if len(stages) > 0:
    if not vytools.utils.exists(stages, items):
      return False
    anchor_copy = copy.deepcopy(anchors)
    if not vytools.stage.build(stages, items, anchor_copy, built, build_level, jobpath=jobpath):
      return False
  return built

def run(original_list, items=None, anchors=None, clean=False, save=False, object_mods=None, jobpath=None, cmd=None, persist=False):
  if items is None: items=ITEMS
  if anchors is None: anchors = {}
  results = {}
  for type_name in original_list:
    anchor_copy = copy.deepcopy(anchors)
    if type_name.startswith('stage:'):
      results[type_name] = vytools.stage.run(type_name, items=items, anchors=anchor_copy, jobpath=jobpath, cmd=cmd)
    elif type_name.startswith('compose:'):
      epid = type_name.replace(':','__')
      results[type_name] = vytools.compose.run(epid, type_name, items=items, anchors=anchor_copy, object_mods=object_mods, jobpath=jobpath, persist=persist)
    elif type_name.startswith('episode:'):
      results[type_name] = vytools.episode.run(type_name, items=items, anchors=anchor_copy, save=save, clean=clean, object_mods=object_mods, jobpath=jobpath, persist=persist)
  return results

def _recursive_list(name, spctop, items, field):
  def list_all_dependencies(name, spc, visited, items):
    if name not in items:
      vytools.printer.print_def(spc+name+'[?]')
    elif name in visited:
      vytools.printer.print_def(spc+name+'[*]')
    else:
      vytools.printer.print_def(spc+name)
      visited.append(name)
      for d in items[name][field]:
        list_all_dependencies(d, spc+spctop, visited, items)
  visited = []
  list_all_dependencies(name, spctop, visited, items)

def get_sorted_images():
  images = {}
  image_list = [image.id for image in vytools.stage.DOCKER_CLIENT.images.list()]
  for image in vytools.stage.DOCKER_CLIENT.images.list():
    labels = image.attrs['Config']['Labels']
    has_image_dependencies = bool(labels and 'tools.vy.images' in labels)
    images[image.id] = {
      'tags':image.tags,
      'depends_on':[],
      'repos':[] if not labels or 'tools.vy.repos' not in labels else labels['tools.vy.repos'].split(','),
      'has_image_dependencies':has_image_dependencies
    }
    if has_image_dependencies:
      depimages = [x.split('=')[-1] for x in labels['tools.vy.images'].split(',')]
      images[image.id]['depends_on'] = [x for x in depimages if x in image_list]
  sorted_image_list = vytools.utils.sort(images.keys(), images)
  return (images, sorted_image_list)

def image_info(original_list,print_dependencies,prefx):
  (images, sorted_image_list) = get_sorted_images()
  for hid in [False, True]:
    for k in sorted_image_list:
      ok = not original_list or any([tag in original_list for tag in images[k]['tags']])
      if ok and images[k]['has_image_dependencies'] == hid:
        nobody_depends_on = '*' if not any([k in i['depends_on'] for i in images.values()]) else ' '
        depends_on_nobody = '-' if not hid else '*' if len(images[k]['depends_on']) == 0 else ' '
        vytools.printer.print_def(prefx+'|{}|{}|'.format(depends_on_nobody,nobody_depends_on) + ','.join(images[k]['tags']))
        if print_dependencies:
          for x in images[k]['repos']: vytools.printer.print_plus(prefx+'     '+x)

  vytools.printer.print_info(prefx+'|-| | = Not built by vy or built before version 0.2.7')
  vytools.printer.print_info(prefx+'|*| | = This image depends on no other images')
  vytools.printer.print_info(prefx+'| |*| = No other vy images depend on this image')

def info(original_list, items=None, list_dependencies=False, list_private=False, list_images=False, expand=False):
  if items is None: items=ITEMS
  CONFIG.info(list_private=list_private)
  dep_spaces = '   '
  for i in original_list:
    if i in items:
      vytools.printer.print_def(i, attrs=['bold'])
      vytools.printer.print_def(json.dumps(items[i], indent=6, sort_keys=True))
      if i.startswith('object:') and expand:
        item = items[i]
        obj,deps = vytools.object.expand(item['data'], item['definition'], items)
        vytools.printer.print_def(json.dumps(obj, indent=6, sort_keys=True))
      if list_dependencies:
        vytools.printer.print_def(dep_spaces + 'dependencies', attrs=['bold'])
        _recursive_list(i, dep_spaces, items, 'depends_on')
        vytools.printer.print_def(dep_spaces + 'dependents', attrs=['bold'])
        _recursive_list(i, dep_spaces, items, 'depended_on')

      repo_versions = vytools.utils.get_repo_versions([i],items)
      sorted_repo_versions = sorted([v for v in repo_versions.values()])
      vytools.printer.print_def('This item depends directly or indirectly on these repositories which are currently at these versions:')
      for x in sorted_repo_versions:
        vytools.printer.print_plus('  '+x)
      if i.startswith('stage:') and list_images:
        vytools.printer.print_def('This item was used to create these images:')
        image_info(vytools.stage.stages(i),list_dependencies,'  ')

    else:
      vytools.printer.print_def('Unknown item {i}'.format(i=i), attrs=['bold'])

  if not original_list:
    vytools.printer.print_def('  Repositories in this context')
    repo_versions = vytools.utils.get_repo_versions(items,items)
    for x in sorted([v for v in repo_versions.values()]):
      vytools.printer.print_plus('    '+x)
  
    if list_images:
      vytools.printer.print_def('  Images on this device')
      image_info([],list_dependencies,'    ')



import subprocess, os, sys, shutil, signal, logging, copy, json
import vytools.utils
import vytools.printer
import vytools.compose
from termcolor import cprint
from vytools.config import ITEMS, CONFIG

global SHUTDOWN
SHUTDOWN = {}
def _shutdown_reset():
  global SHUTDOWN
  SHUTDOWN['path'] = ''
  SHUTDOWN['down'] = []
  SHUTDOWN['logs'] = []
  SHUTDOWN['services'] = []
  
_shutdown_reset()

def stop():
  logs = subprocess.run(SHUTDOWN['logs'], cwd=SHUTDOWN['path'], stdout=subprocess.PIPE)
  subprocess.run(SHUTDOWN['down'], cwd=SHUTDOWN['path'])
  _shutdown_reset()
  return logs

def compose_exit_code(eppath):
  success = True
  try:
    anyzeros = False
    if not os.path.isdir(eppath):
      return False

    # Get services, wish there was a better way...
    services = []
    for s in subprocess.check_output(SHUTDOWN['services'], 
        cwd=SHUTDOWN['path']).decode('ascii').strip().split('\n'):
      count = 1
      while True:
        name = SHUTDOWN['jobid']+'_'+s+'_'+str(count)
        count+=1
        if name not in services:
          break
      services.append(name)

    for service in services:
      try:
        exitcode = subprocess.check_output(['docker', 'container', 'inspect',service,'--format','{{.State.ExitCode}}']).decode('ascii').strip()
      except Exception as exc:
        success = False
        exitcode = '1'
        logging.error('Failed to get exit code for {s}: {e}'.format(s=service, e=exc))
      anyzeros |= int(exitcode) == 0
      logging.info('---- Service '+service+' exited with code '+exitcode)
    return success and anyzeros
  except Exception as exc:
    logging.error('Failed to get exit codes'+str(exc))
    return False

ORIGINAL_SIGINT = signal.getsignal(signal.SIGINT)
def exit_gracefully(signum, frame):
  signal.signal(signal.SIGINT, ORIGINAL_SIGINT) # restore the original signal handler
  # logs = stop()
  #sys.exit(signum) # TODO is this right? pass out signum?

def runpath(epid, jobpath=None):
  if jobpath is None: jobpath = CONFIG.job_path()
  return os.path.join(jobpath,epid) if epid and jobpath else None

def run(epid, compose_name, items=None, anchors=None, clean=False, object_mods=None, jobpath=None, dont_track=None, persist=False):
  if anchors is None: anchors = {}
  if object_mods is None: object_mods = {}
  global SHUTDOWN
  epid = epid.lower()
  if items is None: items = ITEMS
  # TODO test epid, lower case, alphanumeric starts with alpha?
  eppath = runpath(epid,jobpath)
  if not eppath: return False

  if clean:
    try:
      shutil.rmtree(eppath)
    except Exception as exc:
      logging.error('Failed to clean folder {n}'.format(n=eppath))
      return False

  os.makedirs(eppath,exist_ok=True)
  built = []
  # Compile compose files and any volumes
  cbuild = vytools.compose.build(compose_name, items=items, anchors=copy.deepcopy(anchors),
            built=built, build_level=-1, object_mods=object_mods, eppath=eppath) # get components
  if cbuild == False: return False
  if eppath and os.path.exists(eppath):
    with open(os.path.join(eppath, 'vyanchors.json'),'w') as w:
      json.dump(cbuild['anchors'],w,sort_keys=True,indent=1)

  cmd = ['docker-compose'] + cbuild['command'] + ['--project-name', epid]
  cmdup = cmd+['up', '--abort-on-container-exit']
  SHUTDOWN['down'] = cmd + ['down']
  if not persist: SHUTDOWN['down'] += ['--volumes']
  SHUTDOWN['jobid'] = epid
  SHUTDOWN['path'] = eppath
  SHUTDOWN['logs'] = cmd + ['logs']
  SHUTDOWN['services'] = cmd + ['ps','--services']
  try:
    signal.signal(signal.SIGINT, exit_gracefully)
  except Exception as exc:
    logging.warning(str(exc))
    
  with open(os.path.join(eppath,'start.sh'),'w') as w2:
    w2.write(' '.join(cmdup))

  vytools.printer.print_info('Episode Path = '+eppath)
  with open(os.path.join(eppath,'logs.txt'),'w') as logfile:
    proc = subprocess.Popen(cmdup, cwd=eppath, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True)
    for line in proc.stdout:
      vytools.printer.print_def(line.strip(),fp=logfile)
    proc.wait()
  
  compose_exit = compose_exit_code(eppath)
  stop()
  
  repo_versions = vytools.utils.get_repo_versions(cbuild['dependencies'], items)
  stage_versions = cbuild['stage_versions']

  with open(os.path.join(eppath,'logs.txt'),'a') as logfile:
    vytools.printer.print_def('\nThis test was built/run from the following',fp=logfile)
    if dont_track and dont_track in repo_versions:
      this_repo_version = repo_versions[dont_track]
      stage_versions = [sv for sv in stage_versions if sv != this_repo_version]
      vytools.printer.print_def(' - "episode" repositories|versions:',fp=logfile)
      vytools.printer.print_plus('   - '+this_repo_version, fp=logfile)
      del repo_versions[dont_track] # Remove the repository containing this episode

    vytools.printer.print_def(' - "dependency" repositories|versions:', fp=logfile)

    sorted_repo_versions = sorted([v for v in repo_versions.values()])
    for v in sorted_repo_versions:
      vytools.printer.print_plus('   - '+v, fp=logfile)
    vytools.printer.print_def(' - "stage" repositories|versions:', fp=logfile)
    others = [k for k in stage_versions if k not in sorted_repo_versions]
    for v in others:
      vytools.printer.print_plus('   - '+v, fp=logfile)

  return {
    'compose':compose_name,
    'repos':sorted_repo_versions,
    'stage_repos':others,
    'passed':compose_exit and proc.returncode == 0,
    'object_mods':object_mods,
    'anchors':anchors
  }



import os, json, logging, re
import vytools.printer
from pathlib import Path

_FOLDER = os.path.dirname(os.path.realpath(__file__))
_CONFIGFILE = os.path.join(_FOLDER,'vyconfig.json')
_FIELDS = ['contexts','items','jobs','menu','secrets','ssh','scanned']
SEARCHED_REPO_PATHS = {}
ITEMS = {}
KEYMATCH = re.compile("^[A-Za-z_0-9]+$")


def check_folder(field,values):
  success = True
  for v in values:
    if not os.path.isdir(v):
      logging.error('The path "{v}" is not a directory. Create this directory before setting it as your "{f}" directory'.format(v=v,f=field))
      success = False
  return success

class __CONFIG:
  def __init__(self):
    # // 'BITBUCKET_SSH':'/home/nate/.ssh/bitbucketnopassphrase'
    self._data = {}
    try:
      self._data = json.loads(Path(_CONFIGFILE).read_text().strip())
    except:
      pass
  
  def __write(self):
    with open(_CONFIGFILE,'w') as w:
      w.write(json.dumps(self._data))

  def set(self,field,value):
    if field in _FIELDS:
      if field in ['contexts','scanned']:
        if value is None and field == 'scanned':
          self._data[field] = value
        elif check_folder(field, value):
          self._data[field] = [str(Path(v).resolve()) for v in value]
      elif field == 'jobs':
        if check_folder(field, [value]):
          self._data[field] = str(Path(value).resolve())
          ign = os.path.join(self._data[field],'.vyignore')
          if not os.path.exists(ign):
            with open(ign,'w'): pass
      elif field == 'ssh':
        if 'ssh' not in self._data: self._data['ssh'] = {}
        self._data['ssh'].update({k:v for k,v in value.items() if KEYMATCH.search(k)}) # update everything
        self._data['ssh'] = {k:v for k,v in self._data['ssh'].items() if KEYMATCH.search(k) and v != '-'} #remove '-'
      elif field == 'secrets':
        if 'secrets' not in self._data: self._data['secrets'] = {}
        self._data['secrets'].update({k:v for k,v in value.items() if KEYMATCH.search(k)}) # update everything
        self._data['secrets'] = {k:v for k,v in self._data['secrets'].items() if KEYMATCH.search(k) and v != '-'} #remove '-'
      else: # menu, items
        self._data[field] = value
      self.__write()

  def get(self,field):
    return self._data[field] if self._data and field in self._data else None

  def secrets_cmd(self, secret_list, ssh_list):
    cmd = []
    ok = True
    lists = {'secret':secret_list, 'ssh':ssh_list}
    for typ,lst in lists.items():
      for secret in lst:
        if typ == 'secret' and secret in self._data['secrets']:
          cmd += ['--secret','id='+secret+',src='+self._data['secrets'][secret]]
        elif typ == 'ssh' and secret in self._data['ssh']:
          cmd += ['--ssh',secret+'='+self._data['ssh'][secret]]
        else:
          if typ == 'secret':
            secrets_path = self.get('secrets')
            msg = '\n'
            if not secrets_path:
              msg += '  Secrets used by vytools must exist in a "secrets" folder.\n'
              msg += '  Secrets folder(s) can be added via the command line\n'
              msg += '    (e.g. "vytools --secrets path/to/folder/containing/secrets1,path/to/secrets2")\n'
            msg += '  A file named "{s}" containing the secret does not exist in the secrets folders\n'.format(s=secret)
            if secrets_path:
              msg += '    ({f}).'.format(f=','.join(secrets_path))
            logging.error(msg)
          else:
            logging.error('This ssh path for {s} has not been stored in vy, please specify path with vytools (e.g. "vytools --ssh {s}=/path/to/key")'.format(s=secret))
          ok = False
    return (ok, cmd)

  def info(self, list_private=False):
    vytools.printer.print_def('VyTools Configuration & Summary:', attrs=['bold'])
    vytools.printer.print_def('  Configuration saved at: '+_CONFIGFILE)
    vytools.printer.print_def('  Contexts = {s}'.format(s=self.get('contexts')))
    if list_private:
      for ss in ['ssh','secrets']:
        for k,v in self._data.get(ss,{}).items():
          vytools.printer.print_def('  Private file ({t}) {k} @ {v}'.format(t=ss,k=k,v=v))
    jobpath = self.get('jobs')
    if jobpath:
      vytools.printer.print_def('  Jobs at {s}'.format(s=jobpath))
    else:
      vytools.printer.print_def('  Jobs path not specified, specify path with vytools --jobs , (e.g. "vytools --jobs path/to/folder/for/jobs" info)', attrs=['bold'])
    vytools.printer.print_def('  Items found:')
    items = self.get('items')
    typs = {}
    if items:
      for i in items:
        typ = i.split(':')[0]
        if typ not in typs: typs[typ] = 0
        typs[typ] += 1
    for typ in typs:
      vytools.printer.print_def('    - {n} {t} items'.format(t=typ,n=typs[typ]))

  def job_path(self):
    jp = self.get('jobs')
    if not jp:
      logging.error('No jobs path specified ("vytools info")')
      return None
    return jp

CONFIG = __CONFIG()

